//---------------------------------------------------------
// barrera-cv3.cc
//---------------------------------------------------------

#include <unistd.h>            // alarm
#include <condition_variable>  // condition_variable
#include <iostream>            // cout
#include <mutex>               // mutex
#include <string>              // string
#include <thread>              // thread

//---------------------------------------------------------

const int N = 8;

//---------------------------------------------------------

class barrera_t
{
public:
	barrera_t(unsigned _limite): 
		contador(_limite), generacion(0), limite(_limite) {}

	void esperar()
	{
		std::unique_lock<std::mutex> lock(mutex);
		unsigned copia = generacion;
		if (--contador == 0)
		{
			++generacion;
			contador = limite;
			bloqueadas.notify_all();
		}
		else
		{
			bloqueadas.wait(lock, [&]{ return copia != generacion; });
		}
	}

private:
	std::mutex mutex;
	std::condition_variable bloqueadas;
	unsigned contador, generacion, limite;
} barrera(N);

//---------------------------------------------------------

void hebra(int yo)
{
	std::string   antes = std::to_string(yo) +   ": antes\n", 
	            despues = std::to_string(yo) + ": después\n";
	while(true)
	{
		std::cout << antes << std::flush;
		barrera.esperar();
		std::cout << despues << std::flush;
	}
}

//---------------------------------------------------------

int main()
{
	ualarm(150000, 0); // 150ms
	std::thread t[N];
	for (int i = 0; i < N; ++i) t[i] = std::thread(hebra, i);
	for (auto& i: t) i.join();
}

//---------------------------------------------------------

