//---------------------------------------------------------
// semaphore.cc: barrera con semáforos de c++20
//---------------------------------------------------------

#include <unistd.h>     // alarm
#include <atomic>       // atomic
#include <iostream>     // cout endl
#include <mutex>        // mutex
#include <semaphore>    // binary_semaphore
#include <string>       // string
#include <thread>       // thread

//---------------------------------------------------------

const int N = 8;

//---------------------------------------------------------

class barrera_t
{
public:
	explicit barrera_t(unsigned _limite):
		en_espera(0), uso(0), limite(_limite) {}
	
	void esperar()
	{
		unsigned uso_local = uso;
		m.acquire();
		if (++en_espera < limite)
		{
			m.release();
			sem[uso_local].acquire();
		}
		else
		{
			uso = 1 - uso;
			en_espera = 0;
			m.release();
			for (int i = 0; i < N - 1; ++i)
				sem[uso_local].release();
		}
	}

private:
	std::atomic<unsigned> en_espera, uso;
	unsigned limite;
	std::binary_semaphore m{1}, sem[2] = {std::binary_semaphore{0}, std::binary_semaphore{0}};
} barrera(N);

//---------------------------------------------------------

void hebra(int yo)
{
	std::string   antes = std::to_string(yo) +   ": antes\n", 
	            despues = std::to_string(yo) + ": después\n";
	while(true)
	{
		std::cout << antes << std::flush;
		barrera.esperar();
		std::cout << despues << std::flush;
	}
}

//---------------------------------------------------------

int main()
{
	ualarm(150000, 0); // 150ms
	std::thread t[N];
	for (int i = 0; i < N; ++i) t[i] = std::thread(hebra, i);
	for (auto& i: t) i.join();
}

//---------------------------------------------------------

