//---------------------------------------------------------
// spinlock-s2.cc
//---------------------------------------------------------

#include <unistd.h>  // alarm
#include <atomic>    // atomic
#include <chrono>    // us
#include <iostream>  // cout endl
#include <mutex>     // mutex
#include <string>    // string
#include <thread>    // thread

using namespace std::chrono;

//---------------------------------------------------------

const int N = 8;

//---------------------------------------------------------

class barrera_t
{
public:
	explicit barrera_t(unsigned _limit): 
		count(_limit), generation(0), limit(_limit) {}

	void esperar()
	{
		unsigned gen = generation;
		
		if (--count != 0)
		{
			auto espera = limit * 1us;
			while (gen == generation)
				std::this_thread::sleep_for(espera *= 2);
		}
		else
		{
			if (generation.compare_exchange_weak(gen, gen + 1))
				count = limit;
		}
	}

private:
	std::atomic<unsigned> count, generation;
	unsigned limit;
} barrera(N);

//---------------------------------------------------------

void hebra(int yo)
{
	std::string   antes = std::to_string(yo) +   ": antes\n", 
	            despues = std::to_string(yo) + ": después\n";
	while(true)
	{
		std::cout << antes << std::flush;
		barrera.esperar();
		std::cout << despues << std::flush;
	}
}

//---------------------------------------------------------

int main()
{
	ualarm(150000, 0); // 150ms
	std::thread t[N];
	for (int i = 0; i < N; ++i) t[i] = std::thread(hebra, i);
	for (auto& i: t) i.join();
}

//---------------------------------------------------------

